#ifndef ALGO_HEADER
#define ALGO_HEADER

#include <omp.h>
#include <cilk/cilk.h>
#include <cilk/reducer_opadd.h>
#include <cilk/cilk_api.h>
#include <iostream>

#include <math.h>


// SimpleArray class manages data creation and destruction

class SimpleArray {
private:
    double *data;
    unsigned int _size;
public:
    SimpleArray(unsigned int size) : _size(size) 
    {
        data = new double[size];
    }
    ~SimpleArray() 
    { 
        delete [] data; 
    }

    double &operator[](int index) 
    {
        return data[index];
    }

    unsigned int GetSize() 
    {
        return _size;
    }

    void Fill() {
		for(int i=0; i<_size; i++) 
			data[i] = 1.01 + sin((double)i);
    }
};

// Sum is a base class for "+" reduction of an SimpleArray

// Derived classes should override the function 
//      virtual double FindSum(SimpleArray &);
// to introduce other (parallel) algorythms 

class Sum {
protected:
	double *data;
    double operation(double x) {
        return sin(exp(x));
    }
public:
    
	virtual double FindSum(SimpleArray &data) {
		double result(0);
        for(int i=0; i<data.GetSize(); i++)
            result += operation(data[i]);
        return result;
	}

    double RunAndMeasure(SimpleArray &data) {
    	double time(omp_get_wtime());
        double result(FindSum(data));
    	time = omp_get_wtime() - time;                
        std :: cout << " Result: " << result 
            << ", time: " << time << " s" 
            << std :: endl;
        return result;
    }
};

// OpenMPSum class uses pragma's to compute the sum in parallel

class OpenMPSum : public Sum {
public:
    virtual double FindSum(SimpleArray &data) {
        double result(0);
        #pragma omp parallel for reduction (+:result)
        for(int i=0; i<data.GetSize(); i++)
            result += operation(data[i]);
        return result;
    }
};

// CilkForSum class uses cilk_for keyword to compute the sum in parallel
// cilk::reducer_opadd class is used for sync

class CilkForSum : public Sum {
public:
    virtual double FindSum(SimpleArray &data) {
        cilk::reducer_opadd<double> result(0);
        cilk_for(int i=0; i<data.GetSize(); i++)
            result += operation(data[i]);
        return result.get_value();
    }
};

// CilkForNoReducer class also uses cilk_for, but without reducer_opadd
// Note! this class gives erronous results except when being run using one thread only
// Both this class and the one above are implemented to demonstate reducers' performance


class CilkForNoReducerSum : public Sum {
public:
    virtual double FindSum(SimpleArray &data) {
        double result(0);
        cilk_for(int i=0; i<data.GetSize(); i++) {
            result += operation(data[i]);
        }
        return result;
    }
};

// CilkSpawnSum class uses cilk_spawn keyword
// Reduction is performed by function
// _SumComputer(int, int)
// which is run in different threads

class CilkSpawnSum : public Sum {
private:
    int _nThreads;
    cilk::reducer_opadd<double> result;

    void _SumComputer(SimpleArray &data, int i_start, int i_end) {
        double local_result(0);
        for(int i=i_start; i<i_end; i++)
            local_result += operation(data[i]);
        result += local_result;
    }
public:
    CilkSpawnSum(int nThreads) : _nThreads(nThreads), result(0) {}
    virtual double FindSum(SimpleArray &data) {
        for(int i=0; i<_nThreads; i++) {
            cilk_spawn _SumComputer(data, 
                i * data.GetSize() / _nThreads, 
                (i+1) * data.GetSize() / _nThreads);
        }
        cilk_sync;
        return result.get_value();
    }
};

#endif
